#include "jserver.h"
#include "dlfcn.h"

typedef int (*dispatch_t)(mapi m);

typedef struct module_struct
{
    char *name;
    dispatch_t dispatch;
    /* XXX some sort of bitmask for which phases this dispatch handles */
    void *dl;
    int refcount;
    pool p;
    int destroy;
    struct module_struct *next;
} *module;

module modules = NULL;

module module_get(char *name)
{
    module next;

    next = modules;
    while(next != NULL)
    {
        if(j_strcmp(next->name, name) == 0)
            return next;
        next = next->next;
    }
    return NULL;
}

void module_free(module m)
{
    module next, prev;

    prev = next = modules;
    while(next != NULL)
    {
        if(next == m)
        {
            if(next->refcount > 0)
            { /* a thread is still in the module, can't free it yet */
                next->destroy = 1;
                return;
            }

            if(modules == next)
                modules = next->next;
            else
                prev->next = next->next;

            dlclose(next->dl);
            pool_free(next->p);
            return;
        }
        prev = next;
        next = next->next;
    }
    return;
}

void modules_init(xmlnode x)
{
    xmlnode next;
    module mod;
    pool p;
    void *dl;
    dispatch_t dispatch;
    char *dlerr;

    next = xmlnode_get_firstchild(x);
    while(next != NULL)
    {
        if(xmlnode_get_type(next) != NTYPE_TAG)
        {
            next = xmlnode_get_nextsibling(next);
            continue;
        }

        debug(xmlnode_get_name(next));
        mod = module_get(xmlnode_get_name(next));
        if(mod == NULL)
        { /* load the module, create the gloabl list entry */
            dl = dso_load(xmlnode_get_attrib(next,"dso"),NULL,NULL);
            if(dl == NULL)
            {
                warn("module init","load failed");
                next = xmlnode_get_nextsibling(next);
                continue;
            }

            dispatch = (dispatch_t)dlsym(dl,"dispatch");
            dlerr = dlerror();
            if(dlerr != NULL)
            {
                warn("module init err",dlerr);
                next = xmlnode_get_nextsibling(next);
                continue;
            }

            p = pool_new();
            mod = pmalloc(p,sizeof(struct module_struct));
            mod->p = p;
            mod->name = pstrdup(p, xmlnode_get_name(next));
            mod->dl = dl;
            mod->dispatch = dispatch;
            mod->refcount = 0;
            mod->destroy = 0;
            mod->next = modules;
            modules = mod;
        }

        /* send the module it's config data */
        (mod->dispatch)(mapi_new(M_PHASE_CONFIG,mod->p,NULL,jpacket_new(next)));

        next = xmlnode_get_nextsibling(next);
    }

    mod = modules;
    while(mod != NULL)
    {
        /* checking to see if this module we know about is no longer in the config */
        if(xmlnode_get_tag(x, mod->name) == NULL)
            module_free(mod);
        mod = mod->next;
    }
}

mapi mapi_new(int phase, pool p, session s, jpacket pkt)
{
    mapi m;

    if(p == NULL)
        return;

    m = pmalloc(p, sizeof(struct mapi_struct));
    m->phase = phase;
    m->p = p;
    m->s = s;
    m->pkt = pkt;

    return m;
}

void module_dispatch(mapi m)
{
    module next, prev;
    int ret;

    next = modules;
    while(next != NULL)
    {
        /* XXX check this module against this session's list of modules */
        /* XXX check bitmask, maybe this module ignores this phase */
        next->refcount++;
        ret = (next->dispatch)(m);
        next->refcount--;
        switch(ret)
        {
            case M_OK:
                /* XXX store this module on a list in this session's pool */
                break;
            case M_IGNORE_PHASE:
                /* XXX save this knowledge to the modules bitmask so we don't ask the module for this phase again */
                break;
            case M_IGNORE_SESSION:
                break;
        }

        prev = next;
        next = next->next;

        if(prev->destroy == 1)
        {
            module_free(prev);
        }
    }
}
